import enum


class HashableDict(dict):
    def __hash__(self):
        """

        :return:
        """
        return hash(tuple(sorted(self.items())))


class ObjectDict(dict):
    def __init__(self, d=None, **kwargs):
        """

        :param d: input dictionary
        """
        super().__init__()
        data = d or kwargs or {}

        for k, v in data.items():
            if isinstance(v, dict):
                v = ObjectDict(v)
            elif isinstance(v, list):
                v = [
                    ObjectDict(i)
                    if type(i) is dict else i
                    for i in v
                ]

            self[k] = v

    def __dict__(self):
        return {**self}

    def __getattr__(self, name):
        if name in self:
            return self[name]

    def __setattr__(self, name, value):
        if type(value) is dict:
            self[name] = ObjectDict(value)
        else:
            self[name] = value

    def __delattr__(self, name):
        if name in self:
            del self[name]

    @staticmethod
    def normalize(data):
        try:
            if isinstance(data, (list, tuple, set)):
                return [ObjectDict(**r) for r in data]
            return ObjectDict(**data)
        except (TypeError, AttributeError):
            return data


class IntEnum(enum.IntEnum):
    @classmethod
    def to_list(cls):
        return [
            ObjectDict(id=getattr(cls, m).value, label=getattr(cls, m).name)
            for m in cls.__members__
        ]

    def to_dict(self):
        return ObjectDict(id=self.value, label=self.name)

    def __repr__(self):
        return f"<{self.value}: {self.name}>"

    def __str__(self):
        return self.name



# noinspection PyUnresolvedReferences
from flask import current_app as cap

from flaskel.http import httpcode
from flaskel.tester import TestClient
from flaskel.utils import datastruct
from .builder import AppBuilder
from .ext import BASE_EXTENSIONS
from .flaskel import Flaskel
from .standalone import Server
from .version import *
from .wsgi import WSGIFactory



import flask
from flask_limiter import Limiter

from flaskel import cap, httpcode
from flaskel.ext import cfremote


def response_ok(r):
    return httpcode.is_ok(r.status_code)


def response_ko(r):
    return r.status_code != httpcode.TOO_MANY_REQUESTS \
           and httpcode.is_ko(r.status_code)


limiter = Limiter(
    key_func=cfremote.get_remote,
    default_limits=[lambda: cap.config.LIMITER.FAIL],
    default_limits_deduct_when=response_ko
)


def header_whitelist():
    hdr = cap.config.LIMITER.BYPASS_KEY
    value = cap.config.LIMITER.BYPASS_VALUE
    return flask.request.headers.get(hdr) == value


@limiter.request_filter
def _header_whitelist():
    return header_whitelist()


class RateLimit:
    limiter = limiter

    @classmethod
    def slow(cls):
        return cls.limiter.limit(
            lambda: cap.config.LIMITER.SLOW,
            deduct_when=response_ok
        )

    @classmethod
    def medium(cls):
        return cls.limiter.limit(
            lambda: cap.config.LIMITER.MEDIUM,
            deduct_when=response_ok
        )

    @classmethod
    def fast(cls):
        return cls.limiter.limit(
            lambda: cap.config.LIMITER.FAST,
            deduct_when=response_ok
        )

    @classmethod
    def fail(cls):
        return cls.limiter.limit(
            lambda: cap.config.LIMITER.FAIL,
            deduct_when=response_ko
        )


limit_slow = RateLimit.slow()
limit_medium = RateLimit.medium()
limit_fast = RateLimit.fast()
limit_fail = RateLimit.fail()
